/*
** Daedalus (Version 3.1) File: wutil.cpp
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2015 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains Windows specific utilities and operations for Daedalus.
**
** Created: 11/16/2002.
** Last code change: 5/31/2015.
*/

#include <windows.h>
#include <objbase.h>
#include <comdef.h>
#include <comdefsp.h>
#include <shlobj.h>
#include <shfolder.h>
#include <shobjidl.h>
#include <shlwapi.h>
#include <stdio.h>
#include <math.h>

#include "resource.h"
#include "util.h"
#include "graphics.h"
#include "color.h"
#include "threed.h"
#include "maze.h"
#include "draw.h"
#include "daedalus.h"
#include "wdriver.h"


// Constant data

CONST int rgdirDiag[DIRS][DIRS] =
  {{0, 1, 1, 0}, {1, 1, 2, 2}, {3, 2, 2, 3}, {0, 0, 3, 3}};


/*
******************************************************************************
** Windows Routines
******************************************************************************
*/

// Set the window display bitmap to the contents of the active bitmap. This
// usually gets called whenever the active bitmap changes in a major way. The
// window bitmap is designed to avoid flicker in screen updates.

void CopyBitmapToDeviceBitmap()
{
  HDC hdc;

  if (!bm.fColor) {
    if (bm.b.m_x != bm.xa || bm.b.m_y != bm.ya) {
      SelectObject(wi.hdcBmp, wi.hbmpPrev);
      DeleteObject(wi.hbmp);
      wi.hbmp = CreateCompatibleBitmap(wi.hdcBmp, bm.b.m_x, bm.b.m_y);
      if (wi.hbmp == NULL)
        PrintSz_E("Failed to create monochrome display bitmap.");
      wi.hbmpPrev = (HBITMAP)SelectObject(wi.hdcBmp, wi.hbmp);
      bm.xa = bm.b.m_x;
      bm.ya = bm.b.m_y;
    }
    wi.bi.biWidth = bm.xa;
    wi.bi.biHeight = -bm.ya;
    SetDIBitsToDevice(wi.hdcBmp, 0, 0, bm.xa, bm.ya, 0, 0, 0, bm.ya,
      bm.b.m_rgb, (BITMAPINFO *)&wi.bi, DIB_RGB_COLORS);
  } else {
    if (bm.k.m_x != bm.xaC || bm.k.m_y != bm.yaC) {
      SelectObject(wi.hdcBmp, wi.hbmpPrev);
      if (wi.hbmpC != NULL)
        DeleteObject(wi.hbmpC);
      hdc = GetDC(wi.hwnd);
      wi.hbmpC = CreateCompatibleBitmap(hdc, bm.k.m_x, bm.k.m_y);
      if (wi.hbmpC == NULL)
        PrintSz_E("Failed to create color display bitmap.");
      ReleaseDC(wi.hwnd, hdc);
      wi.hbmpPrev = (HBITMAP)SelectObject(wi.hdcBmp, wi.hbmpC);
      bm.xaC = bm.k.m_x;
      bm.yaC = bm.k.m_y;
    }
    wi.biC.biWidth = bm.xaC;
    wi.biC.biHeight = -bm.yaC;
    SetDIBitsToDevice(wi.hdcBmp, 0, 0, bm.xaC, bm.yaC, 0, 0, 0, bm.yaC,
      bm.k.m_rgb, (BITMAPINFO *)&wi.biC, DIB_RGB_COLORS);
  }
}


// The active bitmap being viewed has changed content in a major way. This
// usually means to go update the screen.

void DirtyView()
{
  if (!ws.fNoDirty)
    return;
  CopyBitmapToDeviceBitmap();
  Redraw(wi.hwnd);
}


// Change a pixel in both the main bitmap and on the screen. This updates just
// the pixel in question as opposed to redrawing the whole bitmap.

void BitmapDotCore(int x, int y, bit o)
{
  if (!bm.b.FLegal(x, y))
    return;
  bm.b.Set(x, y, o);
  if (!bm.fColor && ws.fNoDirty)
    SetPixel(wi.hdcBmp, x, y, o ? kvWhite : kvBlack);
}


// Draw a spot, updating bitmaps and screen appropriately. This is called from
// the Set At Dot command and when moving the dot when Drag Is Erase is set.

void BitmapDot(int x, int y)
{
  int x2, y2;

  if (dr.fOnTemp) {

    // Draw a spot on the temporary bitmap.
    if (bm.b2.FNull())
      bm.b2.FBitmapCopy(bm.b);
    if (!bm.b2.FNull()) {
      if (!dr.fBig)
        bm.b2.Set(x, y, dr.fSet);
      else
        bm.b2.Block(x-1, y-1, x+1, y+1, dr.fSet);
    }
  } else if (!bm.fColor) {

    // Draw a spot on the main bitmap and on the screen.
    if (!dr.fBig)
      BitmapDotCore(x, y, dr.fSet);
    else {
      bm.b.Block(x-1, y-1, x+1, y+1, dr.fSet);
      if (!bm.fColor)
        PatBlt(wi.hdcBmp, x-1, y-1, 3, 3, dr.fSet ? WHITENESS : BLACKNESS);
    }
  } else {

    // Draw a spot on the color bitmap and on the screen.
    if (!bm.k.FNull() && bm.k.FLegal(x, y)) {
      if (!dr.fBig) {
        bm.k.Set(x, y, kvSet);
        SetPixel(wi.hdcBmp, x, y, kvSet);
      } else {
        bm.k.Block(x-1, y-1, x+1, y+1, kvSet);
        for (y2 = y-1; y2 <= y+1; y2++)
          for (x2 = x-1; x2 <= x+1; x2++)
            SetPixel(wi.hdcBmp, x2, y2, kvSet);
      }
    }
  }
}


// Change the active bitmap. Switch to having the window show either the main
// bitmap or the color bitmap.

flag FShowColmap(flag fColor)
{
  flag fDirty;

  if (fColor && bm.k.FNull()) {
    if (!bm.k.FColmapGetFromBitmap(bm.b, dr.kvOff, dr.kvOn))
      return fFalse;
  }
  fDirty = (fColor != bm.fColor);
  bm.fColor = fColor;
  CheckMenu(cmdColmap, bm.fColor);
  if (bm.fColor) {
    if (wi.hbmpC != NULL) {
      SelectObject(wi.hdcBmp, wi.hbmpPrev);
      wi.hbmpPrev = (HBITMAP)SelectObject(wi.hdcBmp, wi.hbmpC);
    }
  } else {
    SelectObject(wi.hdcBmp, wi.hbmpPrev);
    wi.hbmpPrev = (HBITMAP)SelectObject(wi.hdcBmp, wi.hbmp);
  }
  if (fDirty)
    DirtyView();
  return fTrue;
}


// Launch an external application or open a Daedalus script, passing it the
// given file. This is called from the Open Documentation command, commands
// on the More Help submenu, and commands on the Run Script submenu.

flag FBootExternal(CONST char *szFile, flag fInternal)
{
  FILE *file;
  char szT[cchSzDef], *pch;

  // The file will be looked for in the directory of the Daedalus executable.
  GetModuleFileName(wi.hinst, szT, cchSzDef);
  for (pch = szT; *pch; pch++)
    ;
  while (pch > szT && *pch != '\\')
    pch--;
  if (*pch == '\\')
    pch++;
  sprintf(SO(pch, szT), "%s", szFile);
  file = FileOpen(szT, "r");
  if (file != NULL)
    goto LDone;

  // Check if the file exists before launching the external app.
  sprintf(S(szT), "%s", szFile);
  file = FileOpen(szT, "r");
  if (file != NULL)
    goto LDone;

  if (!fInternal) {
    sprintf(S(szT), "File '%s' not found.", szFile);
    PrintSz_E(szT);
  }
  return fFalse;

LDone:
  fclose(file);
  if (!fInternal)
    ShellExecute(NULL, NULL, szT, NULL, NULL, SW_SHOW);
  else
    FFileOpen(cmdOpenScript, szT, NULL);
  return fTrue;
}


// Set the delay of the Windows timer event, as controlled with the Repeat
// Delay In Msec setting.

void EnsureTimer(int nDelay)
{
  flag fWantTimer = FWantTimer();

  // Kill the existing timer, if a timer is no longer needed, or the desired
  // delay is different from the existing delay.
  if (ws.lTimer != 0 &&
    (nDelay < 0 || !fWantTimer || (fWantTimer && nDelay != ws.nTimerDelay))) {
    KillTimer(wi.hwndMain, 1);
    ws.lTimer = 0;
  }

  // Create a new timer, if one is needed and there's no active timer.
  if (ws.lTimer == 0 && nDelay >= 0 && fWantTimer)
    ws.lTimer = SetTimer(wi.hwndMain, 1, Max(nDelay, 1), NULL);
  ws.nTimerDelay = nDelay;
}


// Get the number of milliseconds that have passed. Implements the Get Timer
// command and the Timer and Timer2 functions.

ulong GetTimer()
{
  DWORD l;

  if (ws.fTimePause)
    l = ws.timePause;
  else
    l = GetTickCount();
  return l - ws.timeReset;
}


// Pause or unpause the timer. Implements the Pause Timer command.

void PauseTimer(flag f)
{
  DWORD l;

  if (f == ws.fTimePause)
    return;
  ws.fTimePause = f;
  CheckMenu(cmdTimePause, ws.fTimePause);
  if (ws.fTimePause)
    ws.timePause = GetTickCount();
  else {
    l = GetTickCount();
    ws.timeReset += (l - ws.timePause);
  }
}


// For each menu command that can have a check mark by it, determine its state
// and set or clear the check appropriately. This is called after actions that
// may change many settings at once, such as closing certain dialogs.

void RedoMenu(HWND hwnd)
{
  if (!ws.fMenuDirty)
    return;
  ws.fMenuDirty = fFalse;

  CheckMenu(cmdSpree, ws.fSpree);
  CheckMenu(cmdMessage, ws.fIgnorePrint);
  CheckMenu(cmdTimePause, ws.fTimePause);
  CheckMenu(cmdDotShow, dr.fDot);
  CheckMenu(cmdDotWall, dr.fWall);
  CheckMenu(cmdDotTwo, dr.fTwo);
  CheckMenu(cmdDotDrag, dr.fDrag);
  CheckMenu(cmdDotOn, !dr.fSet);
  CheckMenu(cmdDotBig, dr.fBig);
  CheckMenu(cmdDotBoth, dr.fBoth);
  CheckMenu(cmdInsideView, dr.fInside);
  CheckMenu(cmdInsideMap, dr.fMap);
  CheckMenu(cmd3D, dr.f3D);
  CheckMenu(cmdColmap, bm.fColor);
  CheckMenu(cmdPolishMaze, dr.fPolishMaze);
  DrawMenuBar(hwnd);
}


// Get the screen coordinates of the upper left corner of the Daedalus window
// and the pixel size of the window.

void GetWindowPosition(int *xo, int *yo, int *xs, int *ys)
{
  RECT rc;
  POINT pt;

  if (ws.fResizeClient) {

    // Units relative to upper left corner of the drawable or client area.
    *xs = ws.xpClient; *ys = ws.ypClient;
    pt.x = pt.y = 0;
    ClientToScreen(wi.hwnd, &pt);
    *xo = pt.x; *yo = pt.y;
  } else {

    // Units relative to upper left corner of the outer window frame.
    GetWindowRect(wi.hwnd, &rc);
    *xs = rc.right - rc.left; *ys = rc.bottom - rc.top;
    *xo = rc.left; *yo = rc.top;
  }
}


// Convert drawable or client area coordinates to outer window frame
// coordinates, for the upper left corner and size of the Daedalus window.

void AdjustWindowPosition(int *xo, int *yo, int *xs, int *ys)
{
  RECT rc;
  POINT pt;

  pt.x = pt.y = 0;
  ClientToScreen(wi.hwnd, &pt);
  GetWindowRect(wi.hwnd, &rc);
  *xs += (rc.right - rc.left - ws.xpClient);
  *ys += (rc.bottom - rc.top - ws.ypClient);
  *xo += rc.left - pt.x;
  *yo += rc.top - pt.y;
}


// Move the dot or current position one unit. Implements the commands on the
// Move Dot and Move Relative submenus.

void DotMove(int wCmd)
{
  CVector v;
  int nInside = dr.nInside, xold, yold, zold, dold, eold, dnew,
    xsav, ysav, zsav, xcur, ycur, zcur, dcur, din, xinold, yinold, zinold,
    xmax, ymax, dmov, nDot, cMove, zEl, zLo, zHi, i;
  long xin, yin;
  flag fMove, fBreak, fT;
  MSG msg;

  // Free Movement inside display modes: Move or rotate.
  if ((dr.fInside && FBetween(nInside, nInsideFree, nInsideVeryFree)) ||
    dr.fInSmooth) {
    for (cMove = 1 + (!dr.fInSmooth && (wCmd == cmdMove4 || wCmd == cmdMove6 ?
      dr.fBoth : dr.fTwo)); cMove > 0; cMove--) {
      xold = dr.x; yold = dr.y; zold = dr.z;
      xinold = dr.nOffsetX; yinold = dr.nOffsetY; zinold = dr.nOffsetZ;
      switch (wCmd) {
      case cmdMove8: case cmdMoveForward:
      case cmdMove2: case cmdMoveBack:
      case cmdMove7: case cmdMove9: case cmdMove1: case cmdMove3:
        xin = InsideFromA(dr.x) + dr.nOffsetX;
        yin = InsideFromA(dr.y) + dr.nOffsetY;
        din = dr.dir * 90 + dr.nOffsetD;
        i = wCmd >= cmdMove7 && wCmd != cmdMoveBack ? 1 : -1;
        xin -= i * (int)((real)dr.nSpeedXY * RSinD((real)din));
        yin -= i * (int)((real)dr.nSpeedXY * RCosD((real)din));
        dr.x = AFromInside(xin);
        dr.y = AFromInside(yin);
        dr.nOffsetX = (int)(xin - InsideFromA(dr.x));
        dr.nOffsetY = (int)(yin - InsideFromA(dr.y));
        if (wCmd == cmdMove7 || wCmd == cmdMove3)
          goto LFree4;
        else if (wCmd == cmdMove9 || wCmd == cmdMove1)
          goto LFree6;
        break;
      case cmdMove4: LFree4:
        dr.nOffsetD += dr.nSpeedD;
        if (dr.nOffsetD > 45) {
          dr.nOffsetD -= 90;
          DirInc(dr.dir);
        }
        break;
      case cmdMove6: LFree6:
        dr.nOffsetD -= dr.nSpeedD;
        if (dr.nOffsetD <= -45) {
          dr.nOffsetD += 90;
          DirDec(dr.dir);
        }
        break;
      case cmdMoveU:
        if ((dr.f3DNotFree && !dr.fInSmooth) || !dr.f3D) {
          nInside = nInsideSmooth;
          goto LNormal;
        }
        dr.nOffsetZ -= dr.nSpeedZ;
        if (dr.nOffsetZ <= -(dr.zCell >> 1)) {
          dr.nOffsetZ += dr.zCell;
          dr.z -= 1 + dr.fNarrow;
        }
        break;
      case cmdMoveD:
        if ((dr.f3DNotFree && !dr.fInSmooth) || !dr.f3D) {
          nInside = nInsideSmooth;
          goto LNormal;
        }
        dr.nOffsetZ += dr.nSpeedZ;
        if (dr.nOffsetZ > dr.zCell >> 1) {
          dr.nOffsetZ -= dr.zCell;
          dr.z += 1 + dr.fNarrow;
        }
        break;
      case cmdMoveNorth:
      case cmdMoveWest:
      case cmdMoveSouth:
      case cmdMoveEast:
        i = ((wCmd == cmdMoveSouth || wCmd == cmdMoveEast) << 1) |
          (wCmd == cmdMoveWest || wCmd == cmdMoveEast);
        xin = InsideFromA(dr.x) + dr.nOffsetX + xoff[i] * dr.nSpeedXY;
        yin = InsideFromA(dr.y) + dr.nOffsetY + yoff[i] * dr.nSpeedXY;
        dr.x = AFromInside(xin);
        dr.y = AFromInside(yin);
        dr.nOffsetX = (int)(xin - InsideFromA(dr.x));
        dr.nOffsetY = (int)(yin - InsideFromA(dr.y));
        nInside = nInsideFree;
        break;
      case cmdMoveAround:
        dr.dir ^= 2;
        nInside = nInsideFree;
        break;
      }
      if (dr.nEdge == nEdgeTorus) {
        if (!dr.f3D)
          bm.b.Legalize2(&dr.x, &dr.y);
        else {
          while (dr.x < 0)
            dr.x += bm.b.m_x3;
          while (dr.x >= bm.b.m_x3)
            dr.x -= bm.b.m_x3;
          while (dr.y < 0)
            dr.y += bm.b.m_y3;
          while (dr.y >= bm.b.m_y3)
            dr.y -= bm.b.m_y3;
          while (dr.z < 0)
            dr.z += bm.b.m_z3;
          while (dr.z >= bm.b.m_z3)
            dr.z -= bm.b.m_z3;
        }
      }
      if (wCmd == cmdMove8 || wCmd == cmdMove2 || wCmd == cmdMoveU ||
        wCmd == cmdMoveD || wCmd == cmdMoveForward || wCmd == cmdMoveBack ||
        wCmd == cmdMoveNorth || wCmd == cmdMoveWest || wCmd == cmdMoveSouth ||
        wCmd == cmdMoveEast) {
        if (!FUndoIfHitWall(xold, yold, zold, xinold, yinold, zinold, bm.b)) {
          if (dr.fDrag && dr.fOnTemp)
            DotTemp(xold, yold, zold);
        } else
          cMove = 1;
      }
      if (ws.iEventMove2 > 0 && !dr.fInSmooth &&
        (dr.x != xold || dr.y != yold || dr.z != zold))
        FCheckEvent(ws.iEventMove2, dr.x, dr.y, dr.z);
    }
    if (!(dr.fInside && nInside == nInsideVeryFree && !dr.fInSmooth) ||
      (ws.fSpree && !dr.fRedrawAfter))
      Redraw(wi.hwnd);
    return;
  }

  // Draw coordinates: Move or rotate.
  if (ds.fArrow && !dr.fInside) {
    cMove = 1 + (wCmd == cmdMove4 || wCmd == cmdMove6 ? dr.fBoth : dr.fTwo);
    v.Sphere(rDegQuad - (real)ds.theta, (real)ds.phi);
    v.Mul((real)(ds.speedm * cMove));
    switch (wCmd) {
    case cmdMove8:
    case cmdMove7: case cmdMove9:
      ds.hormin += v.NX(); ds.vermin += v.NY(); ds.depmin -= v.NZ();
      if (wCmd == cmdMove7)
        goto LDraw4;
      else if (wCmd == cmdMove9)
        goto LDraw6;
      break;
    case cmdMove2:
    case cmdMove1: case cmdMove3:
      ds.hormin -= v.NX(); ds.vermin -= v.NY(); ds.depmin += v.NZ();
      if (wCmd == cmdMove3)
        goto LDraw4;
      else if (wCmd == cmdMove1)
        goto LDraw6;
      break;
    case cmdMove4: LDraw4:
      ds.theta -= ds.speedr * cMove;
      if (ds.theta < 0)
        ds.theta += nDegMax;
      break;
    case cmdMove6: LDraw6:
      ds.theta += ds.speedr * cMove;
      if (ds.theta >= nDegMax)
        ds.theta -= nDegMax;
      break;
    case cmdMoveU:
      ds.depmin += ds.speedm * cMove;
      break;
    case cmdMoveD:
      ds.depmin -= ds.speedm * cMove;
      break;
    case cmdMoveAround:
      ds.theta += nDegHalf;
      if (ds.theta >= nDegMax)
        ds.theta -= nDegMax;
      break;
    }

    // Rerender line or patch file in memory since draw positioning changed.
    if (ws.fNoDirty) {
      if (!ds.fDidPatch)
        FRenderPerspectiveWire(BFocus(), bm.coor, bm.ccoor);
      else
        FRenderPerspectivePatch(BFocus(), bm.patch, bm.cpatch);
      DirtyView();
    }
    return;
  }

  // Standard case: Change the coordinates of the dot.
LNormal:
  xold = dr.x; yold = dr.y; zold = dr.z; dold = dr.dir; eold = dr.zElev;
  nDot = !dr.fInside ? (1 + dr.fTwo) : (1 + dr.fNarrow);
  dnew = dr.dir; dmov = -1;
  fMove = fFalse;
  switch (wCmd) {
  case cmdMove8: wCmd = dr.fInside ? cmdMoveForward : cmdMoveNorth; break;
  case cmdMove2: wCmd = dr.fInside ? cmdMoveBack    : cmdMoveSouth; break;
  }
  if (!dr.fInside) {
    switch (wCmd) {
    case cmdMove4: wCmd = cmdMoveWest; break;
    case cmdMove6: wCmd = cmdMoveEast; break;
    }
  }
  switch (wCmd) {
  case cmdMoveNorth: dnew = dmov = 0; break;
  case cmdMoveWest:  dnew = dmov = 1; break;
  case cmdMoveSouth: dnew = dmov = 2; break;
  case cmdMoveEast:  dnew = dmov = 3; break;
  case cmdMoveForward: dmov = dnew; break;
  case cmdMoveBack:    dmov = dnew ^ 2; break;
  case cmdMove4: dnew = dr.dir + 1 & DIRS1; break;
  case cmdMove6: dnew = dr.dir - 1 & DIRS1; break;
  case cmdMoveU:
    if (dr.f3D)
      dnew = dmov = 4;
    else {
      if (!dr.fWall || dr.zStep < 0 || !FGetWallVar(dr.x, dr.y, dr.z, &zEl,
        &zLo, &zHi) || !(dr.zElev + (dr.zWall >> 1) <= zLo &&
        dr.zElev + (dr.zWall >> 1) + dr.nSpeedZ > zLo))
        dr.zElev += dr.nSpeedZ;
    }
    break;
  case cmdMoveD:
    if (dr.f3D)
      dnew = dmov = 5;
    else {
      if (!dr.fWall || dr.zStep < 0 || !FGetWallVar(dr.x, dr.y, dr.z, &zEl,
        &zLo, &zHi) || !((dr.zElev >= zHi && dr.zElev - dr.nSpeedZ < zHi) ||
        (dr.zElev >= zEl && dr.zElev - dr.nSpeedZ < zEl)))
        dr.zElev -= dr.nSpeedZ;
      if (dr.zElev < dr.zGround)
        dr.zElev = dr.zGround;
    }
    break;
  case cmdMove1:
    if (dr.fInside) {
      dnew = dr.dir - 1 & DIRS1; dmov = dnew ^ 2;
    } else {
      dnew = rgdirDiag[1][dr.dir]; dmov = DIRS4+1;
    }
    break;
  case cmdMove3:
    if (dr.fInside) {
      dnew = dr.dir + 1 & DIRS1; dmov = dnew ^ 2;
    } else {
      dnew = rgdirDiag[2][dr.dir]; dmov = DIRS4+2;
    }
    break;
  case cmdMove7:
    if (dr.fInside) {
      dnew = dr.dir + 1 & DIRS1; dmov = dnew;
    } else {
      dnew = rgdirDiag[0][dr.dir]; dmov = DIRS4+0;
    }
    break;
  case cmdMove9:
    if (dr.fInside) {
      dnew = dr.dir - 1 & DIRS1; dmov = dnew;
    } else {
      dnew = rgdirDiag[3][dr.dir]; dmov = DIRS4+3;
    }
    break;
  case cmdMoveAround:
    dnew = dr.dir ^ 2;
    break;
  case cmdMoveLeft:
    dnew = dmov = bm.b.PeekWall(dr.x, dr.y, dr.z,
      dr.f3D ? dr.dir2 : dr.dir, dr.f3D, -1);
    break;
  case cmdMoveRight:
    dnew = dmov = bm.b.PeekWall(dr.x, dr.y, dr.z,
      dr.f3D ? dr.dir2 : dr.dir, dr.f3D, 1);
    break;
  case cmdMoveRandom:
    dnew = dmov = bm.b.PeekRandom(dr.x, dr.y, dr.z,
      dr.f3D ? dr.dir2 : dr.dir, dr.f3D);
    break;
  case cmdJump4:
  case cmdJump6:
    dmov = dr.dir + (wCmd == cmdJump4 ? 1 : -1) & DIRS1;
    break;
  }
  if (dmov >= 0) {
    dr.x += xoff3[dmov]*nDot;
    dr.y += yoff3[dmov]*nDot;
    dr.z += zoff3[dmov]*nDot;
  }
  dr.dir2 = dnew;
  if (dnew < DIRS)
    dr.dir = dnew;

  // Handle all the various dot settings.
  if (!bm.fColor) {
    if (!dr.f3D) {
      xmax = bm.b.m_x; ymax = bm.b.m_y;
    } else {
      xmax = bm.b.m_x3; ymax = bm.b.m_y3;
    }
  } else {
    xmax = bm.k.m_x; ymax = bm.k.m_y;
  }
  if (!dr.fInside) {

    // Handle overhead view.
    for (cMove = 0; cMove < 1 + dr.fBoth; cMove++) {
      if (dr.nEdge != nEdgeVoid) {
        if (dr.x < 0)
          dr.x = dr.nEdge == nEdgeTorus ? xmax + dr.x : 0;
        else if (dr.x >= xmax)
          dr.x = dr.nEdge == nEdgeTorus ? dr.x - xmax : xmax-1;
        if (dr.y < 0)
          dr.y = dr.nEdge == nEdgeTorus ? ymax + dr.y : 0;
        else if (dr.y >= ymax)
          dr.y = dr.nEdge == nEdgeTorus ? dr.y - ymax : ymax-1;
      }
      if (!dr.f3D) {

        // Handle overhead view, 2D bitmap.
        if (FHitWall(dr.x, dr.y, dr.z, &dr.zElev)) {
          if (dr.fGlance && dr.x != xold && dr.y != yold &&
            (fT = FHitWall(dr.x, yold, dr.z, NULL)) !=
            FHitWall(xold, dr.y, dr.z, NULL)) {
            if (fT)
              dr.x = xold;
            else
              dr.y = yold;
          } else
            CheckEventMove(ws.iEventWall, xold, yold, dr.z, eold, fFalse);
        } else if (dr.fNoCorner && dr.x != xold && dr.y != yold &&
          FHitWall(dr.x, yold, dr.z, NULL) &&
          FHitWall(xold, dr.y, dr.z, NULL) && !dr.fTwo) {
          dr.x = xold; dr.y = yold;
        }
        if (dr.fDrag) {
          BitmapDot(xold, yold);
          BitmapDot(dr.x, dr.y);
        }
      } else {

        // Handle overhead view, 3D bitmap.
        if (dr.nEdge != nEdgeVoid) {
          if (dr.z < 0)
            dr.z = dr.nEdge == nEdgeTorus ? bm.b.m_z3 + dr.z : 0;
          else if (dr.z >= bm.b.m_z3)
            dr.z = dr.nEdge == nEdgeTorus ? dr.z - bm.b.m_z3 : bm.b.m_z3-1;
        }
        if (FHitWall(dr.x, dr.y, dr.z, &dr.zElev))
          CheckEventMove(ws.iEventWall, xold, yold, zold, eold, fFalse);
        if (dr.fDrag) {
          BitmapDot(bm.b.X2(xold, zold), bm.b.Y2(yold, zold));
          BitmapDot(bm.b.X2(dr.x, dr.z), bm.b.Y2(dr.y, dr.z));
        }
      }
      if (dr.fBoth && cMove == 0) {
        fMove = dr.x != xold || dr.y != yold || dr.z != zold;
        xold = dr.x; yold = dr.y; zold = dr.z;
        if (dmov >= 0) {
          dr.x += xoff3[dmov]*nDot;
          dr.y += yoff3[dmov]*nDot;
          dr.z += zoff3[dmov]*nDot;
        }
      }
    }
    dr.nOffsetX = dr.nOffsetY = dr.nOffsetZ = dr.nOffsetD = 0;
  } else {

    // Handle perspective inside view.
    if (!dr.f3D) {

      // Handle perspective inside view, 2D bitmap.
      if (dr.nEdge == nEdgeStop)
        bm.b.Legalize(&dr.x, &dr.y);
      else if (dr.nEdge == nEdgeTorus)
        bm.b.Legalize2(&dr.x, &dr.y);
    } else {

      // Handle perspective inside view, 3D bitmap.
      if (dr.nEdge == nEdgeStop) {
        if (dr.x < 0)
          dr.x = 0;
        else if (dr.x >= Even(bm.b.m_x3)-nDot)
          dr.x = Odd(bm.b.m_x3)-nDot;
        if (dr.y < 0)
          dr.y = 0;
        else if (dr.y >= Even(bm.b.m_y3)-nDot)
          dr.y = Odd(bm.b.m_y3)-nDot;
        if (dr.z < 0)
          dr.z = 0;
        else if (dr.z >= Odd(bm.b.m_z3))
          dr.z = Odd(bm.b.m_z3)-1;
      } else if (dr.nEdge == nEdgeTorus) {
        while (dr.x < 0)
          dr.x += bm.b.m_x3;
        while (dr.x >= bm.b.m_x3)
          dr.x -= bm.b.m_x3;
        while (dr.y < 0)
          dr.y += bm.b.m_y3;
        while (dr.y >= bm.b.m_y3)
          dr.y -= bm.b.m_y3;
        while (dr.z < 0)
          dr.z += bm.b.m_z3;
        while (dr.z >= bm.b.m_z3)
          dr.z -= bm.b.m_z3;
      }
    }
    if (dr.x != xold || dr.y != yold || dr.z != zold)
      if ((dr.fNarrow && FHitWall((xold + dr.x) >> 1, (yold + dr.y) >> 1,
        (zold + dr.z) >> 1, NULL) || FHitWall(dr.x, dr.y, dr.z, &dr.zElev)))
        CheckEventMove(ws.iEventWall, xold, yold, zold, eold, fFalse);
  }

  // Fire events since the dot has moved.
  if (ws.iEventEdge > 0 && bm.b.FLegalLevel(xold, yold, dr.f3D) &&
    !bm.b.FLegalLevel(dr.x, dr.y, dr.f3D))
    CheckEventMove(ws.iEventEdge, xold, yold, zold, eold, fTrue);
  else if (ws.iEventMove > 0 &&
    (dr.x != xold || dr.y != yold || dr.z != zold))
    CheckEventMove(ws.iEventMove, xold, yold, zold, eold, fTrue);

  // If the dot's positioning has changed...
  fMove |= dr.x != xold || dr.y != yold || dr.z != zold;
  if (dr.zElev != eold && !fMove && ws.fNoDirty &&
    !(dr.fInside && dr.nInside == nInsideVeryFree))
    Redraw(wi.hwnd);
  if (fMove || dr.dir != dold) {
    if (fMove)
      dr.cMove++;

    // Update the screen.
    if (dr.fInside && nInside == nInsideSmooth &&
      NDistance(dr.x, xold, xmax) + NDistance(dr.y, yold, ymax) +
      NDistance(dr.z, zold, bm.b.m_z3) <= 2)
      DotSmoothAnimate(wCmd, xold, yold, zold, dold);
    else if (ws.fNoDirty)
      Redraw(wi.hwnd);

    // If Follow Passages set, keep moving forward until the next junction.
    if (dr.fFollow && fMove && (dr.x == xold || dr.y == yold)) {
      if (dr.z == zold)
        dnew = (dr.x != xold) + ((dr.x > xold || dr.y > yold) << 1);
      else
        dnew = DIRS + (dr.z > zold);
      fBreak = fFalse;
      while (!fBreak && (dr.x != xold || dr.y != yold || dr.z != zold) &&
        (dr.x != dr.x2 || dr.y != dr.y2 || dr.z != dr.z2)) {

        // Pause for the Pixel Display Delay time.
        if (!dr.fInside && dr.fDot)
          for (i = 0; i < ws.nDisplayDelay; i++)
            PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE);

        // Pressing Space will interrupt passage following.
        if (GetAsyncKeyState(VK_SPACE) != 0)
          break;

        if (!dr.f3D) {
          if (!(dr.x > 0 && dr.y > 0 &&
            dr.x < bm.b.m_x-1 && dr.y < bm.b.m_y-1))
            break;
        } else {
          if (!bm.b.FLegalMaze3(dr.x, dr.y, dr.z))
            break;
        }
        xcur = xsav = dr.x; ycur = ysav = dr.y; zcur = zsav = dr.z;
        dcur = dr.dir;
        dnew = bm.b.FollowPassage(&xsav, &ysav, &zsav, dnew, dr.f3D);
        if (FCheckEvent(ws.iEventFollow, xsav, ysav, zsav))
          break;
        dr.x = xsav; dr.y = ysav; dr.z = zsav;
        if (dnew < 0)
          break;
        dr.dir2 = dnew;
        if (dnew < DIRS)
          dr.dir = dnew;
        if (nDot > 1) {
          dr.x += xoff3[dr.dir2];
          dr.y += yoff3[dr.dir2];
          dr.z += zoff3[dr.dir2];
          if (dr.f3D ? !bm.b.FLegalCube(dr.x, dr.y, dr.z) :
            !bm.b.FLegal(dr.x, dr.y)) {
            dr.x = xsav; dr.y = ysav; dr.z = zsav;
            fBreak = fTrue;
          }
        }
        if (dr.x != xcur || dr.y != ycur || dr.z != zcur)
          dr.cMove++;
        if (dr.fDrag)
          BitmapDot(dr.x, dr.y);

        // Update the screen for each new coordinate of the dot.
        if (dr.fInside && nInside == nInsideSmooth) {
          fT = dr.fRedrawAfter; dr.fRedrawAfter &= (ws.iEventFollow > 0);
          DotSmoothAnimate(cmdMove8, xcur, ycur, zcur, dcur);
          dr.fRedrawAfter = fT;
        } else
          Redraw(wi.hwnd);
      }
    }
    if (ws.iEventMove2 > 0 && fMove)
      FCheckEvent(ws.iEventMove2, dr.x, dr.y, dr.z);
  }
}


// Jump the dot or current position by several units, or sidestep by one unit.
// Implements the commands on the Jump Dot submenu.

void DotJump(int wCmd)
{
  CVector v;
  int xold, yold, din, xinold, yinold, xmax, ymax, cMove, i;
  long xin, yin;
  flag fSav;

  // Free Movement inside display modes: Handle sidestepping.
  if ((dr.fInside && FBetween(dr.nInside, nInsideFree, nInsideVeryFree)) ||
    dr.fInSmooth) {
    for (cMove = 1 + !dr.fInSmooth * dr.fTwo; cMove > 0; cMove--) {
      xold = dr.x; yold = dr.y;
      xinold = dr.nOffsetX; yinold = dr.nOffsetY;
      switch (wCmd) {
      case cmdJump4:
      case cmdJump6:
        xin = InsideFromA(dr.x) + dr.nOffsetX;
        yin = InsideFromA(dr.y) + dr.nOffsetY;
        din = dr.dir * 90 + dr.nOffsetD;
        i = wCmd == cmdJump4 ? 1 : -1;
        xin -= i * (int)((real)dr.nSpeedXY * RCosD((real)din));
        yin += i * (int)((real)dr.nSpeedXY * RSinD((real)din));
        dr.x = AFromInside(xin);
        dr.y = AFromInside(yin);
        dr.nOffsetX = (int)(xin - InsideFromA(dr.x));
        dr.nOffsetY = (int)(yin - InsideFromA(dr.y));
        if (dr.nEdge == nEdgeTorus)
          bm.b.Legalize2(&dr.x, &dr.y);
        break;
      }
      if ((wCmd == cmdJump4 || wCmd == cmdJump6)) {
        if (!FUndoIfHitWall(xold, yold, dr.z, xinold, yinold, dr.nOffsetZ,
          bm.b) && dr.fDrag && dr.fOnTemp)
          DotTemp(xold, yold, dr.z);
      }
    }
    if (ws.iEventMove2 > 0 && !dr.fInSmooth && (dr.x != xold || dr.y != yold))
      FCheckEvent(ws.iEventMove2, dr.x, dr.y, dr.z);
    if (!(dr.fInside && dr.nInside == nInsideVeryFree && !dr.fInSmooth))
      Redraw(wi.hwnd);
    return;
  }

  // Draw coordinates: Sidestep or pitch forward/backward.
  if (ds.fArrow && !dr.fInside) {
    cMove = 1 + (wCmd == cmdJump8 || wCmd == cmdJump2 ? dr.fBoth : dr.fTwo);
    v.Sphere((real)ds.theta, (real)ds.phi);
    v.Mul((real)(ds.speedm * cMove));
    switch (wCmd) {
    case cmdJump8:
      ds.phi -= ds.speedr * cMove;
      if (ds.phi < -rDegHalf)
        ds.phi += nDegMax;
      break;
    case cmdJump2:
      ds.phi += ds.speedr * cMove;
      if (ds.phi >= rDegHalf)
        ds.phi -= nDegMax;
      break;
    case cmdJump4:
      ds.hormin -= v.NX(); ds.vermin += v.NY();
      break;
    case cmdJump6:
      ds.hormin += v.NX(); ds.vermin -= v.NY();
      break;
    }

    // Rerender line or patch file in memory since draw positioning changed.
    if (ws.fNoDirty) {
      if (!ds.fDidPatch)
        FRenderPerspectiveWire(BFocus(), bm.coor, bm.ccoor);
      else
        FRenderPerspectivePatch(BFocus(), bm.patch, bm.cpatch);
      DirtyView();
    }
    return;
  }

  // Standard case: Change the coordinates of the dot.
  xold = dr.x; yold = dr.y;
  if (!dr.fInside) {
    switch (wCmd) {
    case cmdJump1:
      dr.x -= zJump; dr.y += zJump; dr.dir = rgdirDiag[1][dr.dir]; break;
    case cmdJump2:
                     dr.y += zJump; dr.dir = 2;                    break;
    case cmdJump3:
      dr.x += zJump; dr.y += zJump; dr.dir = rgdirDiag[2][dr.dir]; break;
    case cmdJump4:
      dr.x -= zJump;                dr.dir = 1;                    break;
    case cmdJump6:
      dr.x += zJump;                dr.dir = 3;                    break;
    case cmdJump7:
      dr.x -= zJump; dr.y -= zJump; dr.dir = rgdirDiag[0][dr.dir]; break;
    case cmdJump8:
                     dr.y -= zJump; dr.dir = 0;                    break;
    case cmdJump9:
      dr.x += zJump; dr.y -= zJump; dr.dir = rgdirDiag[3][dr.dir]; break;
    }
  } else {

    // In perspective inside mode, the Dot Jump commands move one unit at a
    // time, so are implemented in DotMove.
    switch (wCmd) {
    case cmdJump4:
    case cmdJump6:
      // Sidestepping in perspective mode is explicitly covered in DotMove.
      DotMove(wCmd);
      return;
    case cmdJump8:
    case cmdJump2:
      // Shift+move is like normal move except with Follow Passages toggled.
      fSav = dr.fFollow;
      not(dr.fFollow);
      DotMove(wCmd == cmdJump8 ? cmdMove8 : cmdMove2);
      dr.fFollow = fSav;
      return;
    }
  }

  // Handle the dot moving off the edge of the bitmap or level.
  if (dr.f3D) {
    xmax = bm.b.m_x3; ymax = bm.b.m_y3;
  } else if (!bm.fColor) {
    xmax = bm.b.m_x; ymax = bm.b.m_y;
  } else {
    xmax = bm.k.m_x; ymax = bm.k.m_y;
  }
  if (dr.nEdge != nEdgeVoid) {
    if (dr.x < 0)
      dr.x = dr.nEdge == nEdgeTorus ? xmax + dr.x : 0;
    else if (dr.x >= xmax)
      dr.x = dr.nEdge == nEdgeTorus ? dr.x - xmax : xmax-1;
    if (dr.y < 0)
      dr.y = dr.nEdge == nEdgeTorus ? ymax + dr.y : 0;
    else if (dr.y >= ymax)
      dr.y = dr.nEdge == nEdgeTorus ? dr.y - ymax : ymax-1;
  }

  // Fire events since the dot has moved.
  if (ws.iEventEdge > 0 && bm.b.FLegalLevel(xold, yold, dr.f3D) &&
    !bm.b.FLegalLevel(dr.x, dr.y, dr.f3D))
    CheckEventMove(ws.iEventEdge, xold, yold, dr.z, dr.zElev, fTrue);
  else if (ws.iEventMove > 0 && (dr.x != xold || dr.y != yold))
    CheckEventMove(ws.iEventMove, xold, yold, dr.z, dr.zElev, fTrue);
  CenterWithinCell();
  if (dr.x != xold || dr.y != yold)
    Redraw(wi.hwnd);
}


// Shoot an invisible ray of the specified type from the position of the dot
// in the direction it's facing, until whatever sought after is hit.
// Implements the commands on the Zap Dot submenu.

void DotZap(int wCmd, int nParam)
{
  CMazK *c;
  int x, y, dir;
  bit o;

  x = bm.b.X2(dr.x, dr.z); y = bm.b.Y2(dr.y, dr.z);
  o = bm.b.Get(x, y);
  if (FBetween(y, 0, bm.b.m_y-1)) {
    if (x < 0 && dr.dir == 3)
      x = 0;
    else if (x >= bm.b.m_x && dr.dir == 1)
      x = bm.b.m_x-1;
  } else if (FBetween(x, 0, bm.b.m_x-1)) {
    if (y < 0 && dr.dir == 2)
      y = 0;
    else if (y >= bm.b.m_y && dr.dir == 0)
      y = bm.b.m_y-1;
  }
  switch (wCmd) {

  // Destroy wall: Turn off first set pixel encountered in the main bitmap.
  case cmdDotZap:
    while (bm.b.FLegal(x, y) && bm.b.Get(x, y) == o) {
      x += xoff[dr.dir]; y += yoff[dr.dir];
    }
    BitmapDotCore(x, y, o);
    break;

  // Make wall semitransparent: Set the first off pixel encountered in the
  // extra bitmap, where the pixel in the main bitmap is also on.
  case cmdDotZapTrans:
    while (bm.b.FLegal(x, y) &&
      !(bm.b.Get(x, y) && (bm.b3.FNull() || !bm.b3.Get(x, y)))) {
      x += xoff[dr.dir]; y += yoff[dr.dir];
    }
    if (bm.b.FLegal(x, y)) {
      if (bm.b3.FNull()) {
        if (bm.b3.FBitmapCopy(bm.b))
          bm.b3.BitmapOff();
      }
      if (!bm.b3.FNull() && bm.b3.FLegal(x, y)) {
        bm.b3.Set1(x, y);
        if (dr.nTrans == nTransNone)
          dr.nTrans = nTransDefault;
      }
    }
    break;

  // Make unsemitransparent: Turn off the first set pixel encountered in the
  // extra bitmap.
  case cmdDotZapUntrans:
    if (!bm.b3.FNull()) {
      while (bm.b3.FLegal(x, y) && !bm.b3.Get(x, y)) {
        x += xoff[dr.dir]; y += yoff[dr.dir];
      }
      if (bm.b3.FLegal(x, y)) {
        bm.b3.Set0(x, y);
        if (dr.nTrans == nTransNone)
          dr.nTrans = nTransDefault;
      }
    }
    break;

  // ZapTexture operation: For the first set pixel encountered in the main
  // bitmap, set the texture on the side facing the dot.
  case cmdZapTexture:
    while (bm.b.FLegal(x, y) && (!bm.b.Get(x, y) ||
      (!bm.b3.FNull() && bm.b3.Get(x, y)))) {
      x += xoff[dr.dir]; y += yoff[dr.dir];
    }
    if (bm.b.FLegal(x, y)) {
      c = ColmapGetTexture(0);
      if (c != NULL) {
        if (!c->FLegal(x, y)) {
          if (!c->FBitmapSizeSet(bm.b.m_x, bm.b.m_y))
            break;
          c->BitmapOff();
        }
        dir = dr.dir ^ 2;
        c->Set(x, y, SetT(*c, x, y, dir, nParam));
      }
    }
    break;

  default: Assert(fFalse);
  }
}


// Draw a string of text at the appropriate row in the inside display mode.

void DrawTextLine(HDC hdc, CONST char *sz, int cch, int y)
{
  SIZE size;
  KV kvSav1, kvSav2;
  int nSav;

  GetTextExtentPoint(hdc, sz, cch, &size);
  kvSav1 = GetTextColor(hdc); kvSav2 = GetBkColor(hdc);
  SetTextColor(hdc, dr.kvOff); SetBkColor(hdc, dr.kvOn);
  if (dr.kvOff == dr.kvOn)
    nSav = SetBkMode(hdc, TRANSPARENT);
  TextOut(hdc, (ws.xpClient >> 1) - (size.cx >> 1),
    size.cy*y + (y < 0)*ws.ypClient, sz, cch);
  if (dr.kvOff == dr.kvOn)
    SetBkMode(hdc, nSav);
  SetTextColor(hdc, kvSav1); SetBkColor(hdc, kvSav2);
}


// Create a Windows shortcut file, a link pointing to another file. Called
// from FCreateProgramGroup() and FCreateDesktopIcon() to setup the program.

flag FCreateShortcut(CONST char *szDir, CONST char *szName,
  CONST char *szFile, CONST char *szDesc, int nIcon)
{
  IShellLinkA *pisl = NULL;
  IPersistFile *pipf = NULL;
  char sz[cchSzMax], *pch;
  WCHAR wsz[cchSzMax];
  HRESULT hr;
  flag fRet = fFalse;

  // Define the link data
  hr = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER,
    IID_IShellLink, (LPVOID *)&pisl);
  if (FAILED(hr))
    goto LExit;
  GetModuleFileName(wi.hinst, sz, cchSzDef);
  if (nIcon >= 0) {
    hr = pisl->SetIconLocation(sz, nIcon);
    if (FAILED(hr))
      goto LExit;
  }
  for (pch = sz; *pch; pch++)
    ;
  while (pch > sz && *pch != '\\')
    pch--;
  *pch = chNull;
  hr = pisl->SetWorkingDirectory(sz);
  if (FAILED(hr))
    goto LExit;
  sprintf(SO(pch, sz), "\\%s", szFile);
  hr = pisl->SetPath(sz);
  if (FAILED(hr))
    goto LExit;
  if (szDesc == NULL)
    szDesc = szName;
  hr = pisl->SetDescription(szDesc);
  if (FAILED(hr))
    goto LExit;

  // Save it to a link file
  hr = pisl->QueryInterface(IID_IPersistFile, (LPVOID *)&pipf);
  if (FAILED(hr))
    goto LExit;
  sprintf(S(sz), "%s", szDir);
  for (pch = sz; *pch; pch++)
    ;
  sprintf(SO(pch, sz), "\\%s.lnk", szName);
  MultiByteToWideChar(CP_ACP, 0, sz, -1, wsz, cchSzMax);
  hr = pipf->Save(wsz, fTrue);
  if (FAILED(hr))
    goto LExit;

  fRet = fTrue;
LExit:
  if(pipf)
    pipf->Release();
  if(pisl)
    pisl->Release();
  return fRet;
}


// Delete a Windows shortcut file, a link pointing to another file. Called
// from FCreateProgramGroup() to setup the program.

void DeleteShortcut(CONST char *szDir, CONST char *szFile)
{
  char sz[cchSzMax];

  sprintf(S(sz), "%s\\%s.lnk", szDir, szFile);
  _unlink(sz);
}


// Add an icon pointing to the Daedalus executable on the Windows desktop.

flag FCreateDesktopIcon()
{
  char szDir[cchSzMax], szName[cchSzMax];
  LPITEMIDLIST pidl;

  SHGetSpecialFolderLocation(wi.hwnd, CSIDL_DESKTOPDIRECTORY, &pidl);
  SHGetPathFromIDList(pidl, szDir);
  sprintf(S(szName), "%s %s", szDaedalus, szVersion);
  return FCreateShortcut(szDir, szName,
    "daedalus.exe", "Daedalus executable", 0);
}


// Create a Windows program group folder containing icons for Daedalus.

flag FCreateProgramGroup(flag fAll)
{
  char szDir[cchSzMax], szName[cchSzMax], *pch;
  LPITEMIDLIST pidl;
  int iScript;

  // Create program group folder
  SHGetSpecialFolderLocation(wi.hwnd,
    fAll ? CSIDL_COMMON_PROGRAMS : CSIDL_PROGRAMS, &pidl);
  SHGetPathFromIDList(pidl, szDir);
  for (pch = szDir; *pch; pch++)
    ;
  sprintf(SO(pch, szDir), "\\Daedalus");
  if (!CreateDirectory(szDir, NULL)) {
    if (GetLastError() != ERROR_ALREADY_EXISTS)
      return fFalse;
  }

  // Add main shortcuts in folder
  sprintf(S(szName), "%s %s", szDaedalus, szVersion);
  if (!FCreateShortcut(szDir, szName,
    "daedalus.exe", "Daedalus executable", -1))
    return fFalse;
  if (!FCreateShortcut(szDir, "Daedalus documentation",
    "daedalus.htm", NULL, -1))
    return fFalse;
  if (!FCreateShortcut(szDir, "Daedalus scripting", "script.htm", NULL, -1))
    return fFalse;
  if (!FCreateShortcut(szDir, "Daedalus changes", "changes.htm", NULL, -1))
    return fFalse;
  if (!FCreateShortcut(szDir, "Daedalus homepage", "daedalus.url", NULL, 3))
    return fFalse;

  // Delete old shortcuts from folder
  DeleteShortcut(szDir, "Daedalus 2.3");
  DeleteShortcut(szDir, "Daedalus 2.4");
  DeleteShortcut(szDir, "Daedalus 2.5");
  DeleteShortcut(szDir, "Daedalus 3.0");

  // Create scripts subfolder
  while (*pch)
    pch++;
  sprintf(SO(pch, szDir), "\\Scripts");
  if (!CreateDirectory(szDir, NULL)) {
    if (GetLastError() != ERROR_ALREADY_EXISTS)
      return fFalse;
  }

  // Add script shortcuts in subfolder
  for (iScript = 0; iScript <= cmdScriptLast - cmdScript01; iScript++) {
    if (!FCreateShortcut(szDir, rgszShortcut[iScript], rgszScript[iScript],
      NULL, iScript != 5/*Dragonslayer*/ && iScript != 8/*Hunger*/ &&
      iScript != 25/*Squares*/ && iScript != 26/*Mandelbrot*/ &&
      iScript != 27/*Pentris*/ ? 1 : 2))
      return fFalse;
  }

  // Delete old script shortcuts from folder
  DeleteShortcut(szDir, "Carleton Farm Maze");  // Now Carleton Farm Maze #1
  return fTrue;
}


// Add info about Daedalus specific file extensions to the Windows registry.

CONST char *rgszKey[] = {"ds", "d3", "dp", "dw", NULL};
CONST char *rgszValue[] =
  {"Script", "3D Bitmap", "Patch File", "Wireframe File"};

flag FRegisterExtensions()
{
  HKEY hkey;
  char szExe[cchSzMax], szIco[cchSzMax], szKey[cchSzDef], *pch;
  int iKey;

  GetModuleFileName(wi.hinst, szExe, cchSzDef);
  sprintf(S(szIco), "%s,1", szExe);
  for (pch = szExe; *pch; pch++)
    ;
  sprintf(SO(pch, szExe), "%s", " %1");
  for (iKey = 0; rgszKey[iKey]; iKey++) {

    // Define .d? extension and point it to a Daedalus File Type
    sprintf(S(szKey), ".%s", rgszKey[iKey]);
    if (RegCreateKey(HKEY_CLASSES_ROOT, szKey, &hkey) != ERROR_SUCCESS)
      return fFalse;
    sprintf(S(szKey), "Daedalus.%s", rgszKey[iKey]);
    if (RegSetValue(hkey, NULL, REG_SZ, szKey, CchSz(szKey)) != ERROR_SUCCESS)
      return fFalse;
    if (RegDeleteKey(hkey, "shell\\open\\command") == ERROR_SUCCESS) {
      // Daedalus 2.2 and before set this
      if (RegDeleteKey(hkey, "shell\\open") == ERROR_SUCCESS)
        RegDeleteKey(hkey, "shell");
    }
    RegCloseKey(hkey);

    // Set File Type name for .d? extensions
    if (RegCreateKey(HKEY_CLASSES_ROOT, szKey, &hkey) != ERROR_SUCCESS)
      return fFalse;
    sprintf(S(szKey), "Daedalus %s", rgszValue[iKey]);
    if (RegSetValue(hkey, NULL, REG_SZ, szKey, CchSz(szKey)) != ERROR_SUCCESS)
      return fFalse;
    RegCloseKey(hkey);

    // Make .d? extension files be opened by Daedalus
    sprintf(S(szKey), "Daedalus.%s\\shell\\open\\command", rgszKey[iKey]);
    if (RegCreateKey(HKEY_CLASSES_ROOT, szKey, &hkey) != ERROR_SUCCESS)
      return fFalse;
    if (RegSetValue(hkey, NULL, REG_SZ, szExe, CchSz(szExe)) != ERROR_SUCCESS)
      return fFalse;
    RegCloseKey(hkey);

    // Set icon for .d? extension files
    sprintf(S(szKey), "Daedalus.%s\\DefaultIcon", rgszKey[iKey]);
    if (RegCreateKey(HKEY_CLASSES_ROOT, szKey, &hkey) != ERROR_SUCCESS)
      return fFalse;
    if (RegSetValue(hkey, NULL, REG_SZ, szIco, CchSz(szIco)) != ERROR_SUCCESS)
      return fFalse;
    RegCloseKey(hkey);
  }
  return fTrue;
}


// Remove Daedalus specific file extensions from the Windows registry.

flag FUnregisterExtensions()
{
  char szKey[cchSzDef];
  int iKey;

  for (iKey = 0; rgszKey[iKey]; iKey++) {

    // Delete .d? extension pointing to a Daedalus File Type
    sprintf(S(szKey), ".%s", rgszKey[iKey]);
    if (SHDeleteKey(HKEY_CLASSES_ROOT, szKey) != ERROR_SUCCESS)
      return fFalse;

    // Delete icon for .d? extension files
    sprintf(S(szKey), "Daedalus.%s", rgszKey[iKey]);
    if (SHDeleteKey(HKEY_CLASSES_ROOT, szKey) != ERROR_SUCCESS)
      return fFalse;
  }
  return fTrue;
}


// Set a Windows variable to a numeric value.

void DoSetVariableW(int ivos, int n)
{
  int xo, yo, xs, ys;
  flag f = (n != 0);

  switch (ivos) {
  case vosScrollX:
    ws.xScroll = n; SetScrollPos(wi.hwnd, SB_HORZ, ws.xScroll, fTrue);
    break;
  case vosScrollY:
    ws.yScroll = n; SetScrollPos(wi.hwnd, SB_VERT, ws.yScroll, fTrue);
    break;
  case vosSizeX:
  case vosSizeY:
  case vosPositionX:
  case vosPositionY:
  case vosDrawable:
  case vosWindowTop:
    GetWindowPosition(&xo, &yo, &xs, &ys);
    switch (ivos) {
    case vosSizeX:     xs               = n; break;
    case vosSizeY:     ys               = n; break;
    case vosPositionX: xo               = n; break;
    case vosPositionY: yo               = n; break;
    case vosDrawable:  ws.fResizeClient = f; break;
    case vosWindowTop: ws.fWindowTop    = f; break;
    }
    if (ws.fResizeClient)
      AdjustWindowPosition(&xo, &yo, &xs, &ys);
    SetWindowPosition(xo, yo, xs, ys);
    break;
  case vosWindowMenu:
    ws.fWindowMenu = f;
    SetMenu(wi.hwnd, ws.fWindowMenu ? NULL : wi.hmenu);
    break;
  case vosWindowScroll:
    ws.fWindowScroll = f;
    ShowScrollBar(wi.hwnd, SB_BOTH, !ws.fWindowScroll);
    break;

  default:
    PrintSzN_E("Setting Windows variable %d is undefined.", ivos);
  }
}


// Get the value of a Windows variable, which will be stored in a numeric
// return parameter.

int NGetVariableW(int ivos)
{
  int n, nT;
  MSG msg;

  switch (ivos) {
  case vosSizeX:     GetWindowPosition(&nT, &nT, &n,  &nT); break;
  case vosSizeY:     GetWindowPosition(&nT, &nT, &nT, &n);  break;
  case vosPositionX: GetWindowPosition(&n,  &nT, &nT, &nT); break;
  case vosPositionY: GetWindowPosition(&nT, &n,  &nT, &nT); break;
  case vosEventKey:   n = (PeekMessage(&msg, wi.hwnd,
    WM_KEYDOWN, WM_KEYDOWN, PM_NOREMOVE | PM_NOYIELD) != 0); break;
  case vosEventMouse: n = (PeekMessage(&msg, wi.hwnd,
    WM_LBUTTONDOWN, WM_RBUTTONDOWN, PM_NOREMOVE | PM_NOYIELD) != 0); break;
  default:
    PrintSzN_E("Getting Windows variable %d is undefined.", ivos);
  }
  return n;
}


// Execute a Windows operation given parameters.

void DoOperationW(int ioos, CONST char *rgch, int cch, int n1, int n2)
{
  char sz[cchSzOpr], szT[cchSzOpr];

  switch (ioos) {
  case oosMessageIns:
    if (wi.hdcInside != NULL) {
      FormatRgchToSz(rgch, cch, sz, cchSzOpr);
      DrawTextLine(wi.hdcInside, sz, CchSz(sz), n2);
    }
    break;
  case oosTitle:
    FormatRgchToSz(rgch, cch, sz, cchSzOpr);
    if (n2) {
      sprintf(S(szT), "%s %s%s%s",
        ws.szAppName, szVersion, *sz ? " " : "", sz);
    } else
      sprintf(S(szT), "%s", sz);
    SetWindowText(wi.hwnd, szT);
    break;
  case oosBeep:
    MessageBeep((UINT)-1);
    break;
  case oosBeep2:
    Beep(n1, n2);
    break;
  case oosDelay:
    if (n1 >= 0)
      Sleep(n1);
    break;
  case oosSystem:
    SendMessage(wi.hwndMain, WM_SYSCOMMAND, n1 < 0 ? SC_MINIMIZE :
      (n1 == 0 ? SC_RESTORE : SC_MAXIMIZE), 0);
    break;
  case oosMenu:
    ModifyMenu(wi.hmenu, cmdMacro01 + n1 - 1, MF_BYCOMMAND | MF_STRING,
      cmdMacro01 + n1 - 1, rgch);
    break;

  default:
    PrintSzN_E("Windows operation %d is undefined.", ioos);
  }
}


// Execute a simple system specific action.

void SystemHook(int ihos)
{
  switch (ihos) {
  case hosRedraw:
    Redraw(wi.hwnd);
    break;
  case hos3DOff:
    CheckMenu(cmd3D, fFalse);
    break;
  case hos3DOn:
    CheckMenu(cmd3D, fTrue);
    break;
  case hosDlgString:
    DoDialog(DlgString, dlgString);
    break;
  case hosDlgString2:
    DoDialog(DlgString2, dlgString2);
    break;
  case hosDlgString3:
    DoDialog(DlgString3, dlgString3);
    break;

  default:
    PrintSzN_E("Windows hook %d is undefined.", ihos);
  }
}


// Change to the system wait cursor, or restore the previous cursor.

void HourglassCursor(size_t *pcursor, flag fSet)
{
  if (fSet)
    *pcursor = (size_t)SetCursor(LoadCursor(NULL, IDC_WAIT));
  else
    SetCursor((HCURSOR)*pcursor);
}

/* wutil.cpp */
